/* Definitions of target machine for GNU compiler.  Vxworks PowerPC 64bit
   version.
   Copyright (C) 2017 Free Software Foundation, Inc.
   Contributed by Douglas B Rupp

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3, or (at your option) any later
version.

GCC is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING3.  If not see
<http://www.gnu.org/licenses/>.  */

#undef TARGET_64BIT
#define TARGET_64BIT 1

#undef DEFAULT_ABI
#define DEFAULT_ABI ABI_ELFv2

/* The VxWorks 7 post SR0600 environment is powerpc-linux based. Consequently,
   much of this section comes from linux64.h */

#if TARGET_VXWORKS7 >= 2

#ifdef HAVE_LD_NO_DOT_SYMS
/* New ABI uses a local sym for the function entry point.  */
/* Currently VxWorks does not use this new ABI (ABI_ELFv2). Left here in case
   this changes in the future. */
/* extern int dot_symbols;
   #undef DOT_SYMBOLS
   #define DOT_SYMBOLS dot_symbols
   */
#endif

/* For now VxWorks does not use local symbols for the function entry point. */
#undef DOT_SYMBOLS
#define DOT_SYMBOLS 0

#define TARGET_USES_LINUX64_OPT 1
/* Enforce a single code model for now.  */
/* #ifdef HAVE_LD_LARGE_TOC
   #undef TARGET_CMODEL
   #define TARGET_CMODEL rs6000_current_cmodel
   #define SET_CMODEL(opt) rs6000_current_cmodel = opt
   #else
   #define SET_CMODEL(opt) do {} while (0)
   #endif
*/
#undef TARGET_CMODEL
#define TARGET_CMODEL CMODEL_MEDIUM

#undef  PROCESSOR_DEFAULT64
#define PROCESSOR_DEFAULT64 PROCESSOR_PPCE6500

#undef  RS6000_ABI_NAME
#define RS6000_ABI_NAME "linux"

/* 64-bit PowerPC Linux always has a TOC.  */
#undef  TARGET_TOC
#define TARGET_TOC              1

/* Some things from sysv4.h we don't do when 64 bit.  */
#undef  OPTION_RELOCATABLE
#define OPTION_RELOCATABLE      0
#undef  OPTION_EABI
#define OPTION_EABI             0
#undef  OPTION_PROTOTYPE
#define OPTION_PROTOTYPE        0
#undef RELOCATABLE_NEEDS_FIXUP
#define RELOCATABLE_NEEDS_FIXUP 0

/* PowerPC64 Linux word-aligns FP doubles when -malign-power is given.  */
#undef  ADJUST_FIELD_ALIGN
#define ADJUST_FIELD_ALIGN(FIELD, TYPE, COMPUTED) \
  (rs6000_special_adjust_field_align_p ((TYPE), (COMPUTED))             \
   ? 128                                                                \
   : (TARGET_64BIT                                                      \
      && TARGET_ALIGN_NATURAL == 0                                      \
      && TYPE_MODE (strip_array_types (TYPE)) == DFmode)                \
   ? MIN ((COMPUTED), 32)                                               \
   : (COMPUTED))

/* PowerPC64 Linux increases natural record alignment to doubleword if
   the first field is an FP double, only if in power alignment mode.  */
#undef  ROUND_TYPE_ALIGN
#define ROUND_TYPE_ALIGN(STRUCT, COMPUTED, SPECIFIED)                   \
  ((TARGET_64BIT                                                        \
    && (TREE_CODE (STRUCT) == RECORD_TYPE                               \
        || TREE_CODE (STRUCT) == UNION_TYPE                             \
        || TREE_CODE (STRUCT) == QUAL_UNION_TYPE)                       \
    && TARGET_ALIGN_NATURAL == 0)                                       \
   ? rs6000_special_round_type_align (STRUCT, COMPUTED, SPECIFIED)      \
   : MAX ((COMPUTED), (SPECIFIED)))

/* Use the default for compiling target libs.  */
#ifdef IN_TARGET_LIBS
#undef TARGET_ALIGN_NATURAL
#define TARGET_ALIGN_NATURAL 1
#endif

/* Indicate that jump tables go in the text section.  */
#undef  JUMP_TABLES_IN_TEXT_SECTION
#define JUMP_TABLES_IN_TEXT_SECTION TARGET_64BIT

/* The linux ppc64 ABI isn't explicit on whether aggregates smaller
   than a doubleword should be padded upward or downward.  You could
   reasonably assume that they follow the normal rules for structure
   layout treating the parameter area as any other block of memory,
   then map the reg param area to registers.  i.e. pad upward.
   Setting both of the following defines results in this behavior.
   Setting just the first one will result in aggregates that fit in a
   doubleword being padded downward, and others being padded upward.
   Not a bad idea as this results in struct { int x; } being passed
   the same way as an int.  */
#define AGGREGATE_PADDING_FIXED TARGET_64BIT
#define AGGREGATES_PAD_UPWARD_ALWAYS 0

/* Specify padding for the last element of a block move between
   registers and memory.  FIRST is nonzero if this is the only
   element.  */
#define BLOCK_REG_PADDING(MODE, TYPE, FIRST) \
  (!(FIRST) ? PAD_UPWARD : targetm.calls.function_arg_padding (MODE, TYPE))

#undef  TOC_SECTION_ASM_OP
#define TOC_SECTION_ASM_OP \
  (TARGET_64BIT                                         \
   ? "\t.section\t\".toc\",\"aw\""                      \
   : "\t.section\t\".got\",\"aw\"")

#undef  MINIMAL_TOC_SECTION_ASM_OP
#define MINIMAL_TOC_SECTION_ASM_OP \
  (TARGET_64BIT                                         \
   ? "\t.section\t\".toc1\",\"aw\""                     \
   : (flag_pic                                          \
      ? "\t.section\t\".got2\",\"aw\""                  \
      : "\t.section\t\".got1\",\"aw\""))

/* Must be at least as big as our pointer type.  */
#undef  SIZE_TYPE
#define SIZE_TYPE (TARGET_64BIT ? "long unsigned int" : "unsigned int")

#undef  PTRDIFF_TYPE
#define PTRDIFF_TYPE (TARGET_64BIT ? "long int" : "int")

#undef  WCHAR_TYPE
#define WCHAR_TYPE (TARGET_64BIT ? "int" : "long int")
#undef  WCHAR_TYPE_SIZE
#define WCHAR_TYPE_SIZE 32

#undef  RS6000_MCOUNT
#define RS6000_MCOUNT "_mcount"

/* #ifdef __powerpc64__ */
/* _init and _fini functions are built from bits spread across many
   object files, each potentially with a different TOC pointer.  For
   that reason, place a nop after the call so that the linker can
   restore the TOC pointer if a TOC adjusting call stub is needed.  */
#if DOT_SYMBOLS
#define CRT_CALL_STATIC_FUNCTION(SECTION_OP, FUNC)      \
  asm (SECTION_OP "\n"                                  \
"       bl ." #FUNC "\n"                                \
"       nop\n"                                          \
"       .previous");
#else
#define CRT_CALL_STATIC_FUNCTION(SECTION_OP, FUNC)      \
  asm (SECTION_OP "\n"                                  \
"       bl " #FUNC "\n"                                 \
"       nop\n"                                          \
"       .previous");
#endif
/* #endif */

/* FP save and restore routines.  */
#undef  SAVE_FP_PREFIX
#define SAVE_FP_PREFIX (TARGET_64BIT ? "._savef" : "_savefpr_")
#undef  SAVE_FP_SUFFIX
#define SAVE_FP_SUFFIX ""
#undef  RESTORE_FP_PREFIX
#define RESTORE_FP_PREFIX (TARGET_64BIT ? "._restf" : "_restfpr_")
#undef  RESTORE_FP_SUFFIX
#define RESTORE_FP_SUFFIX ""

/* Dwarf2 debugging.  */
#undef  PREFERRED_DEBUGGING_TYPE
#define PREFERRED_DEBUGGING_TYPE DWARF2_DEBUG

/* This is how to declare the size of a function.  */
#undef  ASM_DECLARE_FUNCTION_SIZE
#define ASM_DECLARE_FUNCTION_SIZE(FILE, FNAME, DECL)                    \
  do                                                                    \
    {                                                                   \
      if (!flag_inhibit_size_directive)                                 \
        {                                                               \
          fputs ("\t.size\t", (FILE));                                  \
          if (TARGET_64BIT && DOT_SYMBOLS)                              \
            putc ('.', (FILE));                                         \
          assemble_name ((FILE), (FNAME));                              \
          fputs (",.-", (FILE));                                        \
          rs6000_output_function_entry (FILE, FNAME);                   \
          putc ('\n', (FILE));                                          \
        }                                                               \
    }                                                                   \
  while (0)

/* Return nonzero if this entry is to be written into the constant
   pool in a special way.  We do so if this is a SYMBOL_REF, LABEL_REF
   or a CONST containing one of them.  If -mfp-in-toc (the default),
   we also do this for floating-point constants.  We actually can only
   do this if the FP formats of the target and host machines are the
   same, but we can't check that since not every file that uses
   the macros includes real.h.  We also do this when we can write the
   entry into the TOC and the entry is not larger than a TOC entry.  */

#undef  ASM_OUTPUT_SPECIAL_POOL_ENTRY_P
#define ASM_OUTPUT_SPECIAL_POOL_ENTRY_P(X, MODE)                        \
  (TARGET_TOC                                                           \
   && (GET_CODE (X) == SYMBOL_REF                                       \
       || (GET_CODE (X) == CONST && GET_CODE (XEXP (X, 0)) == PLUS      \
           && GET_CODE (XEXP (XEXP (X, 0), 0)) == SYMBOL_REF)           \
       || GET_CODE (X) == LABEL_REF                                     \
       || (GET_CODE (X) == CONST_INT                                    \
           && GET_MODE_BITSIZE (MODE) <= GET_MODE_BITSIZE (Pmode))      \
       || (GET_CODE (X) == CONST_DOUBLE                                 \
           && ((TARGET_64BIT                                            \
                && (TARGET_MINIMAL_TOC                                  \
                    || (SCALAR_FLOAT_MODE_P (GET_MODE (X))              \
                        && ! TARGET_NO_FP_IN_TOC)))                     \
               || (!TARGET_64BIT                                        \
                   && !TARGET_NO_FP_IN_TOC                              \
                   && SCALAR_FLOAT_MODE_P (GET_MODE (X))                \
                   && BITS_PER_WORD == HOST_BITS_PER_INT)))))

/* Select a format to encode pointers in exception handling data.  CODE
   is 0 for data, 1 for code labels, 2 for function pointers.  GLOBAL is
   true if the symbol may be affected by dynamic relocations.  */
#undef  ASM_PREFERRED_EH_DATA_FORMAT
#define ASM_PREFERRED_EH_DATA_FORMAT(CODE, GLOBAL) \
  (TARGET_64BIT || flag_pic                                             \
   ? (((GLOBAL) ? DW_EH_PE_indirect : 0) | DW_EH_PE_pcrel               \
      | (TARGET_64BIT ? DW_EH_PE_udata8 : DW_EH_PE_sdata4))             \
   : DW_EH_PE_absptr)

/* Unclear where this gets defined on the powerpc-linux target, but appears
   to be in effect */
#undef DEFAULT_PCC_STRUCT_RETURN
#define DEFAULT_PCC_STRUCT_RETURN 1


/* The VxWorks 7 pre SR0600 environment on 64bit PowerPC is Diab based.  */

#else /* TARGET_VXWORKS7 >= 2 */

#undef DOT_SYMBOLS
#define DOT_SYMBOLS 0

#undef DEFAULT_PCC_STRUCT_RETURN
#define DEFAULT_PCC_STRUCT_RETURN 1

/* And enforce a single code model for now.  */

#undef TARGET_CMODEL
#define TARGET_CMODEL CMODEL_MEDIUM
#endif /* TARGET_VXWORKS7 >= 2 */
