/* Support for ARM EABI unwinding in VxWorks Downloadable Kernel Modules.
   Copyright (C) 2017-2019 Free Software Foundation, Inc.

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any
   later version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Under Section 7 of GPL version 3, you are granted additional
   permissions described in the GCC Runtime Library Exception, version
   3.1, as published by the Free Software Foundation.

   You should have received a copy of the GNU General Public License and
   a copy of the GCC Runtime Library Exception along with this program;
   see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
   <http://www.gnu.org/licenses/>.  */

#if defined(__vxworks) && !defined (__RTP__)
#if TOOL_FAMILY==llvm

/*
   For 7.2 (llvm based) ARM the symbols __gnu_Unwind_Find_exidx and
   __cxa_type_match are not in the kernel image, so we supply them here.
   The latter is a dummy function, since it is not used by C or Ada.  If
   C++ is ported to this target, then this may have to be rethought.
*/

#include <private/moduleLibP.h>

typedef struct
{
    UINT32 fnoffset;
    UINT32 content;
} __EIT_entry;

extern __EIT_entry __exidx_start;
extern __EIT_entry __exidx_end;

__EIT_entry *__gnu_Unwind_Find_exidx (void *pc, int *nrec)
{
  __EIT_entry *pstart = 0;
  __EIT_entry *pend = 0;

  if (_func_moduleExidxGet != NULL)
    {
      _func_moduleExidxGet (pc,
			    (void *) &__exidx_start, (void *) &__exidx_end,
			    (void **) &pstart, (void **) &pend
			   );
    }

  if (!pstart)
    {
      pstart = &__exidx_start;
      pend = &__exidx_end;
    }

  *nrec = pend - pstart;

  return pstart;
}

void *__cxa_type_match () {return (void *) 0;}
#else

/*
   Pre 7.2 (gnu based) Vxworks for ARM uses __gnu_Unwind_Find_exidx to
   retrieve the exception table for downloadable kernel modules.  As
   those modules are only partially linked, the linker won't generate
   __exidx_start|end, but the two symbols are still used in alternate
   paths from unwind-arm-common.inc.
*/

void *__exidx_start __attribute__((__visibility__ ("hidden")));
void *__exidx_end __attribute__((__visibility__ ("hidden")));
#endif
#endif
